# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for handling the generation of images for the training of
tensorflow classification models.
"""

import logging
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from keras_preprocessing.image import DataFrameIterator, DirectoryIterator, Iterator
from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.data_preparation.annotation_handler import AnnotationHandler
from tensorflow.keras.preprocessing.image import ImageDataGenerator

from mlcvzoo_tf_classification.annotation_utils import annotation_list_to_dataframe
from mlcvzoo_tf_classification.configuration import Config, NetConfig, TrainConfig
from mlcvzoo_tf_classification.const import (
    DataGeneratorConfig,
    DataSubsets,
    ImageDataFrameColumns,
    ImageDataFromDirectory,
)

logger = logging.getLogger(__name__)


def setup_generators(
    net_config: NetConfig, configuration: Config, number_classes: int
) -> Tuple[Iterator, Iterator]:
    """
    Sets up data generators based on the model's configuration

    Returns:
        A training data generator
        A validation data generator
    """
    img_gen = Generators(
        net_config=net_config,
        train_config=configuration.train_config,
        string_replacement_map=configuration.string_replacement_map,
        number_classes=number_classes,
    )
    return img_gen.get_train_generator(), img_gen.get_val_generator()


class Generators:
    """
    A wrapper for the DataIterators offered by keras.preprocessing
    How data will be generated is specified in the respective model config file.
    """

    def __init__(
        self,
        net_config: NetConfig,
        train_config: TrainConfig,
        number_classes: int,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> None:
        self.img_size = net_config.input_shape[:2]
        self.number_classes = number_classes

        self.train_config = train_config

        self.string_replacement_map: Optional[Dict[str, str]] = string_replacement_map

        self._build_train_generator()
        self._build_val_generator()
        self._build_test_generator()

    def _build_train_generator(self) -> None:
        if self.train_config.train_generator_config.flow_from_directory is not None:
            self.train_generator = self.build_generator(
                subset=DataSubsets.TRAINING,
                directory=self.train_config.train_generator_config.flow_from_directory.directory,
            )
        elif self.train_config.train_generator_config.flow_from_dataframe is not None:
            self.annotation_handler = AnnotationHandler(
                configuration=self.train_config.train_generator_config.flow_from_dataframe.annotation_handler_config,
                yaml_config_path=self.train_config.train_generator_config.flow_from_dataframe.annotation_handler_config_path,
                string_replacement_map=self.string_replacement_map,
            )

            self.train_generator = self.build_generator(
                subset=DataSubsets.TRAINING,
                ann_file_list=self.train_config.train_generator_config.flow_from_dataframe.ann_file_list,
            )

            log_message = (
                "ANNOTATION HANDLER CONFIG: "
                f"{self.train_config.train_generator_config.flow_from_dataframe.annotation_handler_config}\n"
                "ANNOTATION HANDLER CONFIG PATH: "
                f"{self.train_config.train_generator_config.flow_from_dataframe.annotation_handler_config_path}"
            )

            logger.info(log_message)

    def _build_val_generator(self) -> None:
        if self.train_config.val_generator_config.flow_from_directory is not None:
            self.val_generator = self.build_generator(
                subset=DataSubsets.VALIDATION,
                directory=self.train_config.val_generator_config.flow_from_directory.directory,
            )
        elif self.train_config.val_generator_config.flow_from_dataframe is not None:
            self.val_generator = self.build_generator(
                subset=DataSubsets.VALIDATION,
                ann_file_list=self.train_config.val_generator_config.flow_from_dataframe.ann_file_list,
            )

    def _build_test_generator(self) -> None:
        if self.train_config.test_generator_config.flow_from_directory is not None:
            self.test_generator = self.build_generator(
                subset=DataSubsets.TEST,
                directory=self.train_config.test_generator_config.flow_from_directory.directory,
            )
        elif self.train_config.test_generator_config.flow_from_dataframe is not None:
            self.test_generator = self.build_generator(
                subset=DataSubsets.TEST,
                ann_file_list=self.train_config.test_generator_config.flow_from_dataframe.ann_file_list,
            )

    def get_train_generator(self) -> Iterator:
        """Returns the generator for training"""
        return self.train_generator

    def get_val_generator(self) -> Iterator:
        """Returns the generator for validation"""
        return self.val_generator

    def get_test_generator(self) -> Iterator:
        """Returns the generator for test"""
        return self.test_generator

    def build_generator(
        self,
        subset: str,
        directory: Optional[str] = None,
        ann_file_list: Optional[List[str]] = None,
    ) -> Iterator:
        """
        Takes a parameter which indicates the subset of the data
        (Training, Validation or Test, see const.DataSubsets),
        and the directory or annotation file, which holds the data.
        Returns an iterator for the specified subset with the specified data.
        Note that the iterator for the training subset is more configurable than
        the other two subset generators.

        Args:
            subset: String that indicates processed subset one of DataSubsets,
            directory: Optional path where the data for the generator is located,
            ann_file_list: Optional list that holds the data

        Returns:
            Iterator, either DirectoryIterator or DataFrameIterator depending on data source

        Raises:
            ValueError, if the provided subset is not in DataSubsets or none of the data sources
             is given.
        """
        logger.info(
            " - annotation file: %s,\n - data subset: %s", ann_file_list, subset
        )

        if subset == DataSubsets.TRAINING:
            # Base train/validation generator
            data_generator = ImageDataGenerator(
                rescale=self.train_config.rescale,
                validation_split=DataGeneratorConfig.VALIDATION_SPLIT,
                featurewise_center=self.train_config.featurewise_center,
                featurewise_std_normalization=self.train_config.featurewise_std_normalization,
                rotation_range=self.train_config.rotation_range,
                width_shift_range=self.train_config.width_shift_range,
                height_shift_range=self.train_config.height_shift_range,
                horizontal_flip=self.train_config.horizontal_flip,
                vertical_flip=self.train_config.vertical_flip,
                zoom_range=self.train_config.zoom_range,
            )

        elif subset in (DataSubsets.VALIDATION, DataSubsets.TEST):
            # to avoid augmentation of data in validation split
            data_generator = ImageDataGenerator(
                rescale=self.train_config.rescale,
                validation_split=DataGeneratorConfig.VALIDATION_SPLIT,
            )
        else:
            raise ValueError(
                f"The subset '{subset}' is not support, "
                f"use either of: {DataSubsets.get_values_as_list(class_type=DataSubsets)}"
            )

        if ann_file_list is not None:
            logger.info(" - source: dataframe -> df generator")
            return self._build_df_generator(
                data_generator=data_generator, ann_file_list=ann_file_list
            )
        if directory is not None:
            logger.info(" - source: directory (%s) -> dir generator", directory)
            return self._build_dir_generator(
                data_generator=data_generator, directory=directory
            )
        raise ValueError(
            "Either provide data for parameter 'ann_file_list' or 'directory'"
        )

    def _build_df_generator(
        self, data_generator: ImageDataGenerator, ann_file_list: List[str]
    ) -> DataFrameIterator:

        annotations: List[BaseAnnotation] = []

        for csv_file_path in ann_file_list:
            annotations.extend(
                self.annotation_handler.parse_annotations_from_csv(
                    csv_file_path=csv_file_path
                )
            )

        train_df = annotation_list_to_dataframe(annotations=annotations)

        logger.info(
            " => created dataframe of shape: %s, with columns: %s",
            train_df.shape,
            train_df.columns,
        )

        one_hot_df = self._create_one_hot_encodings(train_df=train_df)
        train_df = train_df.join(one_hot_df)

        logger.info(
            "dataframe columns \n %s \n and column types \n %s",
            train_df.columns,
            train_df.dtypes,
        )

        logger.info("build generator from dataframe")

        df_generator = data_generator.flow_from_dataframe(
            dataframe=train_df,
            directory=None,
            x_col=ImageDataFrameColumns.IMAGE_PATH,
            y_col=list(train_df.iloc[:, 2:].columns),
            subset=None,
            class_mode=ImageDataFrameColumns.CLASS_MODE,
            target_size=self.img_size,
            color_mode=self.train_config.color_mode,
            shuffle=self.train_config.shuffle,
            batch_size=self.train_config.batch_size,
            seed=self.train_config.seed,
            save_to_dir=self.train_config.save_to_dir,
            save_prefix=self.train_config.save_prefix,
            save_format=self.train_config.save_format,
            follow_links=self.train_config.follow_links,
            interpolation=self.train_config.interpolation,
        )

        return df_generator

    def _create_one_hot_encodings(self, train_df: pd.DataFrame) -> pd.DataFrame:
        labels = []

        for i in range(len(train_df)):
            one_hot_enc = np.zeros(self.number_classes)
            class_id = int(train_df.loc[i, ImageDataFrameColumns.CLASS_ID])
            one_hot_enc[class_id] = 1
            labels.append(one_hot_enc)

        labels_as_array = np.asarray(labels)

        one_hot_df = pd.DataFrame(labels_as_array)

        return one_hot_df

    def _build_dir_generator(
        self, data_generator: ImageDataGenerator, directory: str
    ) -> DirectoryIterator:

        logger.info("build generator from directory")

        dir_generator = data_generator.flow_from_directory(
            directory=directory,
            target_size=self.img_size,
            classes=None,
            subset=None,
            class_mode=ImageDataFromDirectory.CLASS_MODE,
            color_mode=self.train_config.color_mode,
            shuffle=self.train_config.shuffle,
            batch_size=self.train_config.batch_size,
            seed=self.train_config.seed,
            save_to_dir=self.train_config.save_to_dir,
            save_prefix=self.train_config.save_prefix,
            save_format=self.train_config.save_format,
            follow_links=self.train_config.follow_links,
            interpolation=self.train_config.interpolation,
        )
        return dir_generator



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
This configuration class is for building python objects from the configuration file.
The respective config fields are parsed via the related component from config_builder module.
"""

from typing import Dict, Optional

import related
from mlcvzoo_base.configuration.detector_configs import DetectorConfig
from mlcvzoo_base.configuration.replacement_config import ReplacementConfig

from mlcvzoo_tf_classification.configuration import Config, TrainConfig
from mlcvzoo_tf_classification.const import LossTypes, OptimizerTypes


@related.mutable(strict=True)
class CustomBlockTrainConfig(TrainConfig):
    """
    Here the hyper parameters for training are extracted.
    """

    optimizer: str = related.StringField(default=OptimizerTypes.SGD)


@related.mutable(strict=True)
class CustomBlockConfig(Config):
    """
    Here the model specific training configuration is extracted and the respective
    detailed config class as listed above is called.
    """

    train_config: CustomBlockTrainConfig = related.ChildField(
        cls=CustomBlockTrainConfig
    )

    base_config: DetectorConfig = related.ChildField(
        cls=DetectorConfig, default=DetectorConfig()
    )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for enumerating options for different configurations for different parameters.
The constant classes listed here are e.g. used in data generators to
limit the options in configuration.
"""

from mlcvzoo_base.configuration.structs import BaseType


class OptimizerTypes(BaseType):
    """Constants for supported Optimizer Types"""

    # other optimizers: https://keras.io/api/optimizers/
    ADAM = "Adam"
    SGD = "SGD"


class LossTypes(BaseType):
    """Constants for supported loss Types"""

    # other losses: https://keras.io/losses/
    CATEGORICAL_CROSSENTROPY = "categorical_crossentropy"


class DataSubsets(BaseType):
    """
    Data subsets for differentiating setups of the data generators
    in Generator class in image_generator.py
    """

    TRAINING = "training"
    VALIDATION = "validation"
    TEST = "test"


class DataGeneratorConfig(BaseType):
    """
    Validation split is not used in our implementation as we offer an interface
    by the net_config.yaml files where a user can define separate directories for
    training, validation and test.

    It is only a necessary value for initializing the ImageDataGenerator from
    tensorflow.keras.preprocessing.image.
    """

    VALIDATION_SPLIT = 0.0


class ImageDataFrameColumns(BaseType):
    """
    Constants for dataframe column names for loading annotation from file into
    dataframe, which is the interface of the keras generator from dataframe
    """

    # the constants are variables of the keras image data generator function flow_from_dataframe
    # the CLASS_ID is for creating the class label column in the used dataframe and the
    #  respective parameter 'classes' in the flow_from_dataframe function
    # IMAGE_PATH is the name of the column in the dataframe which holds the respective
    #  information about the image paths and is used in the 'x_col' parameter of flow_from_dataframe
    # CLASS_MODE is for setting the respective parameter in flow_from_dataframe which describes the
    #  mode of how the label information is given
    # keras api description of the parameters classes, x_col and class_mode can be found here:
    # https://www.tensorflow.org/api_docs/python/tf/keras/preprocessing/image/ImageDataGenerator
    CLASS_ID = "class_id"
    IMAGE_PATH = "image_path"
    CLASS_MODE = "raw"  # one of "binary", "categorical", "input", "multi_output", "raw", sparse" or None


class ImageDataFromDirectory(BaseType):
    """Constants for the class mode when loading data from directory"""

    # class_mode is a variable of the keras image data generator function flow_from_directory
    # keras api description of class mode can be found here:
    # https://www.tensorflow.org/api_docs/python/tf/keras/preprocessing/image/ImageDataGenerator
    CLASS_MODE = (
        "categorical"  # one of "categorical", "binary", "sparse", "input", or None
    )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
The CustomBlockModel configures the model for training and inference.
The respective model is compiled from the provided configuration file and net class.
Input data classes are parsed via the AnnotationClassMapper.
Data flow is regulated via Generators.
"""

import logging
import os
import typing
from typing import Dict, Optional

import numpy as np
import tensorflow as tf
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from tensorflow.keras.layers import Conv2D, Dense, Dropout, Flatten, Input, MaxPooling2D

from mlcvzoo_tf_classification.base_model import BaseModel, ImageType
from mlcvzoo_tf_classification.custom_block.configuration import CustomBlockConfig

logger = logging.getLogger(__name__)


class CustomBlockModel(
    BaseModel[CustomBlockConfig],
):
    """
    The model is compiled from the respective net class and configuration file.
    Input details are parsed using the AnnotationClassMapper.
    """

    def __init__(
        self,
        from_yaml: str,
        configuration: Optional[CustomBlockConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ):
        BaseModel.__init__(
            self,
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )

    def _init_inference_model(self) -> None:
        """
        Initializes the networks architecture and sets the instance's attribute self.net
        to a keras model which is initialized with the first and last layers of
        the defined architecture.

        Idea of using building blocks is inspired by VGG16.
        See respective arXiv paper: https://arxiv.org/pdf/1409.1556.pdf
        """

        # 1st block
        input1 = Input(
            shape=self.configuration.net_config.input_shape
        )  # (100, 100, 3))
        conv1_1 = Conv2D(
            filters=32,
            kernel_size=3,
            activation="relu",
            kernel_initializer="he_uniform",
            padding="same",
        )(input1)
        maxpool_1 = MaxPooling2D(pool_size=2)(conv1_1)
        drop_1 = Dropout(0.2)(maxpool_1)  # learns slower, but breaks in at end

        # 2nd block
        conv2_1 = Conv2D(
            filters=64,
            kernel_size=3,
            activation="relu",
            kernel_initializer="he_uniform",
            padding="same",
        )(drop_1)
        maxpool_2 = MaxPooling2D(pool_size=2)(conv2_1)
        drop_2 = Dropout(0.2)(maxpool_2)

        # 3rd block
        conv3_1 = Conv2D(
            filters=128,
            kernel_size=3,
            activation="relu",
            kernel_initializer="he_uniform",
            padding="same",
        )(drop_2)
        maxpool_3 = MaxPooling2D(pool_size=2)(conv3_1)
        drop_3 = Dropout(0.2)(maxpool_3)

        flatten = Flatten()(drop_3)
        dense_1 = Dense(units=128, activation="relu", kernel_initializer="he_uniform")(
            flatten
        )
        drop_4 = Dropout(0.5)(dense_1)
        dense_2 = Dense(units=self.mapper.num_classes, activation="softmax")(drop_4)

        self.net = tf.keras.Model(inputs=input1, outputs=dense_2, name="custom_model")

        if os.path.isfile(self.configuration.net_config.model_path):
            # Where to add weights into the custom model?
            self.restore(checkpoint_path=self.configuration.net_config.model_path)

    def store(self, checkpoint_path: str) -> None:
        BaseModel.store(self, checkpoint_path)

    def restore(self, checkpoint_path: str) -> None:
        BaseModel.restore(self, checkpoint_path)

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[CustomBlockConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> CustomBlockConfig:
        return typing.cast(
            CustomBlockConfig,
            create_basis_configuration(
                configuration_class=CustomBlockConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @staticmethod
    def preprocess_data(input_data: ImageType) -> ImageType:
        return input_data



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Definition of the Config that is used to configure the models of the
mlcvzoo_tf_classification package.
"""
import logging
from typing import Dict, List, Optional, Union

import related
from config_builder.BaseConfigClass import BaseConfigClass
from mlcvzoo_base.api.configuration import Configuration
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import ReductionMappingConfig

from .const import LossTypes, OptimizerTypes

logger = logging.getLogger(__name__)


@related.mutable(strict=True)
class NetConfig(BaseConfigClass):
    """
    Here the base parameters for the model configuration are extracted.
    """

    model_path: str = related.StringField()

    build_custom_net: bool = related.BooleanField()

    input_shape: List[int] = related.SequenceField(cls=int)

    save_weights_only: Optional[bool] = related.BooleanField(
        required=False, default=True
    )

    def check_values(self) -> bool:
        success: bool = True

        if self.input_shape is not None:
            success = success and len(self.input_shape) == 3
            if not success:
                logger.error(
                    "\nCheck for attribute 'input_shape' failed\n"
                    "condition: len(input_shape) == 3, found input_shape=%s"
                    % self.input_shape
                )

        return success


@related.mutable(strict=True)
class InferenceConfig(BaseConfigClass):
    """
    Here the parameters for model inference are extracted.
    """

    top: Optional[int] = related.IntegerField(required=False, default=None)
    score_threshold: Optional[float] = related.FloatField(required=False, default=None)

    reduction_class_mapping: Optional[ReductionMappingConfig] = related.ChildField(
        cls=ReductionMappingConfig, required=False, default=None
    )

    def check_values(self) -> bool:
        success: bool = True

        if self.top is not None:
            success = success and self.top >= 1
            if not success:
                logger.error(
                    "\nCheck for attribute 'top' failed\n"
                    "condition: top >= 1, found top=%s" % self.top
                )

        if self.score_threshold is not None:
            success = success and 0.0 <= self.score_threshold <= 1.0
            if not success:
                logger.error(
                    "\nCheck for attribute 'top' failed\n"
                    "condition: 0.0 <= score_threshold <= 1.0, found score_threshold=%s"
                    % self.score_threshold
                )

        return success


@related.mutable(strict=True)
class FlowFromDirectoryConfig(BaseConfigClass):
    """
    Here the parameters for the training data are extracted, when it
    is available in a directory.
    """

    directory: str = related.StringField(default="")


@related.mutable(strict=True)
class FlowFromDataframeConfig(BaseConfigClass):
    """
    Here the parameters for the training data are extracted, when it
    is available in a specific file.
    """

    ann_file_list: List[str] = related.SequenceField(cls=str)
    annotation_handler_config_path: str = related.StringField(default="")
    annotation_handler_config: Optional[AnnotationHandlerConfig] = related.ChildField(
        cls=AnnotationHandlerConfig, required=False, default=None
    )


@related.mutable(strict=True)
class GeneratorConfig(BaseConfigClass):
    """
    Here the parameter for data source is extracted. Note that we have
    exclusive options here.
    """

    # mutually exclusive fields
    flow_from_directory: Optional[FlowFromDirectoryConfig] = related.ChildField(
        cls=FlowFromDirectoryConfig, default=None, required=False
    )

    flow_from_dataframe: Optional[FlowFromDataframeConfig] = related.ChildField(
        cls=FlowFromDataframeConfig, default=None, required=False
    )


@related.mutable(strict=True)
class ModelCheckpointConfig(BaseConfigClass):
    """
    Here the parameters for ModelCheckpoint callback are extracted.
    The parameters follow the definition of tensorflow. For detailed
    information on how to set the parameters have a look at their documentation:
    https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/ModelCheckpoint

    Note that the parameter save_weights_only is listed under NetConfig class.
    """

    # specify directory where to store trained models
    work_dir: str = related.StringField()

    monitor: Optional[str] = related.StringField(required=False, default="val_loss")
    verbose: Optional[int] = related.IntegerField(required=False, default=1)
    save_best_only: Optional[bool] = related.BooleanField(required=False, default=False)
    mode: Optional[str] = related.StringField(required=False, default="auto")
    save_freq: Optional[Union[str, int]] = related.StringField(
        required=False, default="epoch"
    )
    initial_value_threshold: Optional[float] = related.FloatField(
        required=False, default=None
    )

    def check_values(self) -> bool:
        return self.verbose in [0, 1] and self.mode in ["auto", "min", "max"]


@related.mutable(strict=True)
class TrainConfig(BaseConfigClass):
    """
    Here the parameters for the data generators and training hyper-parameters
    are extracted.
    """

    batch_size: int = related.IntegerField()
    epochs: int = related.IntegerField()

    metrics: List[str] = related.SequenceField(cls=str)

    rotation_range: int = related.IntegerField()
    horizontal_flip: bool = related.BooleanField()
    zoom_range: float = related.FloatField()
    rescale: bool = related.BooleanField()

    train_generator_config: GeneratorConfig = related.ChildField(cls=GeneratorConfig)
    val_generator_config: GeneratorConfig = related.ChildField(cls=GeneratorConfig)
    test_generator_config: GeneratorConfig = related.ChildField(cls=GeneratorConfig)

    model_checkpoint_config: ModelCheckpointConfig = related.ChildField(
        cls=ModelCheckpointConfig
    )

    featurewise_center: Optional[bool] = related.BooleanField(
        required=False, default=False
    )
    featurewise_std_normalization: Optional[bool] = related.BooleanField(
        required=False, default=False
    )
    vertical_flip: Optional[bool] = related.BooleanField(required=False, default=True)
    width_shift_range: Optional[float] = related.FloatField(required=False, default=0.2)
    height_shift_range: Optional[float] = related.FloatField(
        required=False, default=0.2
    )

    seed: Optional[int] = related.IntegerField(required=False, default=None)

    color_mode: str = related.StringField(default="rgb")
    shuffle: bool = related.BooleanField(default=False)

    # specify a directory where to store augmented images
    save_to_dir: Optional[str] = related.StringField(required=False, default=None)
    save_prefix: Optional[str] = related.StringField(required=False, default="")
    save_format: Optional[str] = related.StringField(required=False, default="png")
    follow_links: Optional[bool] = related.BooleanField(required=False, default=False)
    interpolation: Optional[str] = related.StringField(
        required=False, default="nearest"
    )

    loss: str = related.StringField(default=LossTypes.CATEGORICAL_CROSSENTROPY)
    optimizer: str = related.StringField(default=OptimizerTypes.ADAM)
    learning_rate: float = related.FloatField(default=0.001)
    # momentum optional for training
    momentum: Optional[float] = related.FloatField(default=0.9)

    def check_values(self) -> bool:
        return (
            self.optimizer
            in OptimizerTypes.get_values_as_list(class_type=OptimizerTypes)
            and self.loss in LossTypes.get_values_as_list(class_type=LossTypes)
            and self.color_mode in ["grayscale", "rgb", "rgba"]
            and self.save_format in ["png", "jpeg"]
            and self.interpolation in ["nearest", "bilinear", "bicubic"]
        )


@related.mutable(strict=True)
class Config(BaseConfigClass, Configuration):
    """
    Here the parameter groups are extracted and the respective detailed
    config classes as listed above are called.
    """

    mutual_attribute_map: Dict[str, List[str]] = {
        "GeneratorConfig": ["flow_from_directory", "flow_from_dataframe"]
    }

    net_config: NetConfig = related.ChildField(cls=NetConfig)

    class_mapping: ClassMappingConfig = related.ChildField(ClassMappingConfig)

    inference_config: InferenceConfig = related.ChildField(cls=InferenceConfig)

    train_config: TrainConfig = related.ChildField(cls=TrainConfig)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for handling annotation utility methods that are used across the
mlcvzoo_tf_classification package.
"""

from typing import Any, Dict, List

import pandas as pd
from mlcvzoo_base.api.data.annotation import BaseAnnotation

from mlcvzoo_tf_classification.const import ImageDataFrameColumns


def annotation_list_to_dataframe(annotations: List[BaseAnnotation]) -> pd.DataFrame:
    """
    Build a pandas dataframe object out of a given list of annotations.

    Args:
        annotations: List of BaseAnnotation that should be transformed to a dataframe

    Returns:
        A pandas.DataFrame representation of the input annotations

    """

    annotation_dict: Dict[str, List[Any]] = dict()

    annotation_dict[ImageDataFrameColumns.CLASS_ID] = []
    annotation_dict[ImageDataFrameColumns.IMAGE_PATH] = []

    # TODO: both possibilities to use either classifications or max bounding_boxes

    for annotation in annotations:
        for bounding_box in annotation.get_bounding_boxes(include_segmentations=True):

            # TODO: may change class id to global class id induced by mapping
            annotation_dict[ImageDataFrameColumns.CLASS_ID].append(
                bounding_box.class_id
            )
            annotation_dict[ImageDataFrameColumns.IMAGE_PATH].append(
                annotation.image_path
            )

    annotation_df: pd.DataFrame = pd.DataFrame(annotation_dict)
    annotation_df[ImageDataFrameColumns.CLASS_ID] = annotation_df[
        ImageDataFrameColumns.CLASS_ID
    ].astype(
        "string"
    )  # int64, category
    annotation_df[ImageDataFrameColumns.IMAGE_PATH] = annotation_df[
        ImageDataFrameColumns.IMAGE_PATH
    ].astype("string")

    return annotation_df



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for handling the implementation of a generic base model that wraps tensorflow
classification models.
"""

import logging
import os
from abc import ABC, abstractmethod
from typing import Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union, cast

import numpy as np
import tensorflow as tf
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.interfaces import NetBased, Trainable
from mlcvzoo_base.api.model import ClassificationModel
from nptyping import Int, NDArray, Shape
from tensorflow.keras.callbacks import ModelCheckpoint
from tensorflow.keras.optimizers import SGD, Adam
from tensorflow.keras.preprocessing.image import array_to_img, img_to_array, load_img

from mlcvzoo_tf_classification.const import OptimizerTypes
from mlcvzoo_tf_classification.custom_block.configuration import CustomBlockConfig
from mlcvzoo_tf_classification.image_generator import setup_generators
from mlcvzoo_tf_classification.xception.configuration import XceptionConfig

logger = logging.getLogger(__name__)

ConfigurationType = TypeVar("ConfigurationType", CustomBlockConfig, XceptionConfig)

ImageType = NDArray[Shape["Height, Width, Any"], Int]


class BaseModel(
    ClassificationModel[ConfigurationType, Union[str, ImageType]],
    NetBased[tf.keras.Model],
    Trainable,
    ABC,
    Generic[ConfigurationType],
):
    """
    Generic base model that wraps tensorflow classification models. It provides the functionality
    for handling the configuration, prediction and training of models.
    """

    def __init__(
        self,
        from_yaml: str,
        configuration: Optional[ConfigurationType] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ):
        self.net: Optional[tf.keras.Model] = None
        self.configuration: ConfigurationType = self.create_configuration(  # type: ignore
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )
        ClassificationModel.__init__(
            self,
            unique_name=os.path.basename(from_yaml),
            configuration=self.configuration,
            init_for_inference=True,
            mapper=AnnotationClassMapper(
                class_mapping=self.configuration.class_mapping,
                reduction_mapping=self.configuration.inference_config.reduction_class_mapping,
            ),
        )
        NetBased.__init__(
            self,
            net=self.net,
        )
        Trainable.__init__(self)

    def store(self, checkpoint_path: str) -> None:
        if self.net is None:
            raise ValueError(
                "In order to store a checkpoint, the net attribute has"
                "to be initialized!"
            )

        if self.configuration.net_config.save_weights_only:
            self.net.save_weights(checkpoint_path)
        else:
            tf.keras.models.save_model(self.net, filepath=checkpoint_path)

    def restore(self, checkpoint_path: str) -> None:

        if self.net is None:
            raise ValueError(
                "In order to restore a checkpoint, the net attribute has"
                "to be initialized!"
            )

        if self.configuration.net_config.save_weights_only:
            self.net.load_weights(checkpoint_path)
        else:
            self.net = tf.keras.models.load_model(filepath=checkpoint_path)

    @property
    def num_classes(self) -> int:
        return self.mapper.num_classes

    def get_classes_id_dict(self) -> Dict[int, str]:
        return self.mapper.annotation_class_id_to_model_class_name_map

    @staticmethod
    @abstractmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[ConfigurationType] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ConfigurationType:
        raise NotImplementedError(
            "Must be implemented by sub-class: create_configuration(...)."
        )

    @staticmethod
    @abstractmethod
    def preprocess_data(input_data: ImageType) -> ImageType:
        """
        Preprocesses the given input data

        Args:
            input_data: The input data to preprocess

        Returns:
            The preprocessed data
        """
        raise NotImplementedError(
            "Must be implemented by sub-class: preprocess_data(...)."
        )

    def predict(
        self, data_item: Union[str, ImageType]
    ) -> Tuple[Union[str, ImageType], List[Classification]]:
        """
        Predicts a class for the given data_item.

        Arguments:
            data_item: Input data. It could be:
                - A string to the image file
                - A Numpy array representation of the image
        Returns:
            The data_item and also a list of predictions which length depends
            on the configuration of the model. That is to consider the top x classes
            with the highest scores or only classes with a score not less than the
            configured score value. If these are not set the model will output
            prediction scores for each class.
        """

        assert self.net is not None

        predictions: List[Classification] = []
        target_size = tuple(self.configuration.net_config.input_shape[:2])

        if isinstance(data_item, str):
            image = load_img(
                path=data_item,
                grayscale=False,
                color_mode="rgb",
                target_size=None,
                interpolation="nearest",
            )

            image = image.resize(target_size)
            data_array = img_to_array(image)
            data_array = np.expand_dims(data_array, axis=0)

        elif isinstance(data_item, np.ndarray):
            image = array_to_img(x=data_item)
            image = image.resize(target_size)
            data_array = img_to_array(image)
            data_array = np.expand_dims(data_array, axis=0)

        else:
            # This should be impossible
            logger.warning("Cannot predict on data_item of type: %s", type(data_item))
            return data_item, predictions

        # preprocessing according to  Xception implementation in keras
        data = self.preprocess_data(input_data=data_array)

        # predict labels on data, returns a Tensor with shape (1,nb_classes)
        net_predictions = self.net(data)

        # sort indices by score
        ids_by_score = tf.argsort(net_predictions, axis=1, direction="DESCENDING")
        ids_by_score = tf.keras.backend.eval(ids_by_score)[0]

        for i, _ in enumerate(net_predictions[0]):
            class_id = ids_by_score[i]

            score = tf.cast(net_predictions[0][class_id], float, name=None)

            if (
                self.configuration.inference_config.score_threshold is not None
                and score >= self.configuration.inference_config.score_threshold
            ):
                continue

            class_identifiers = (
                self.mapper.map_model_class_id_to_output_class_identifier(
                    class_id=class_id
                )
            )

            model_class_identifier = ClassIdentifier(
                class_id=class_id,
                class_name=self.mapper.map_annotation_class_id_to_model_class_name(
                    class_id=class_id
                ),
            )

            for class_identifier in class_identifiers:
                predictions.append(
                    Classification(
                        class_identifier=class_identifier,
                        model_class_identifier=model_class_identifier,
                        score=score,
                    )
                )

        if self.configuration.inference_config.top is not None:
            predictions = predictions[: self.configuration.inference_config.top]

        return data_item, predictions

    def train(self) -> None:
        """
        Trains the model specified in self.net with the parameters given in self.configuration.
        """

        assert self.net is not None

        train_generator, validation_generator = setup_generators(
            net_config=self.configuration.net_config,
            configuration=self.configuration,
            number_classes=self.mapper.num_classes,
        )

        if train_generator is not None:
            # set optimizer
            if self.configuration.train_config.optimizer == OptimizerTypes.SGD:
                optimizer = SGD(
                    learning_rate=self.configuration.train_config.learning_rate,
                    momentum=self.configuration.train_config.momentum,
                )
            elif self.configuration.train_config.optimizer == OptimizerTypes.ADAM:
                optimizer = Adam(
                    learning_rate=self.configuration.train_config.learning_rate
                )
            else:
                optimizer = None

            if optimizer is not None:

                # compile model
                self.net.compile(
                    optimizer=optimizer,
                    loss=self.configuration.train_config.loss,
                    metrics=list(self.configuration.train_config.metrics),
                )

                # train the model
                logger.info(
                    "train params: \n"
                    "- epochs: %s,\n"
                    "- batch size: %s,\n"
                    "- validation generator: %s,\n"
                    "- number of val samples: %s, \n"
                    "- number of train samples: %s",
                    self.configuration.train_config.epochs,
                    self.configuration.train_config.batch_size,
                    validation_generator,
                    validation_generator.n,
                    train_generator.n,
                )

                steps_per_epoch = int(
                    train_generator.n / self.configuration.train_config.batch_size
                )
                validation_steps = int(
                    validation_generator.n / self.configuration.train_config.batch_size
                )

                checkpoint_path = os.path.join(
                    self.configuration.train_config.model_checkpoint_config.work_dir,
                    "cp-{epoch:04d}.ckpt",
                )

                callback = ModelCheckpoint(
                    filepath=checkpoint_path,
                    monitor=self.configuration.train_config.model_checkpoint_config.monitor,
                    verbose=self.configuration.train_config.model_checkpoint_config.verbose,
                    save_best_only=self.configuration.train_config.model_checkpoint_config.save_best_only,
                    save_weights_only=self.configuration.net_config.save_weights_only,
                    mode=self.configuration.train_config.model_checkpoint_config.mode,
                    save_freq=self.configuration.train_config.model_checkpoint_config.save_freq,
                    options=None,
                    initial_value_threshold=self.configuration.train_config.model_checkpoint_config.initial_value_threshold,
                )

                self.net.fit(
                    train_generator,
                    callbacks=[callback],
                    epochs=self.configuration.train_config.epochs,
                    steps_per_epoch=steps_per_epoch,
                    validation_data=validation_generator,
                    validation_steps=validation_steps,
                )

            else:
                logger.error(
                    "Could not run training, no optimizer could be initialized"
                )

        else:
            logger.warning(
                "Training data generator is None. Training will not be executed"
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mlcvzoo_tf_classification',
 'mlcvzoo_tf_classification.custom_block',
 'mlcvzoo_tf_classification.xception']

package_data = \
{'': ['*']}

install_requires = \
['keras>=2.6,<3.0',
 'mlcvzoo_base>=4.0,<5.0',
 'numpy>=1.19.2,!=1.19.5',
 'pandas<=1.3.3',
 'protobuf<=3.20',
 'related-mltoolbox>=1.0,<2.0',
 'tensorflow-cpu>=2.6,<3.0',
 'yaml-config-builder>=6.0,<7.0']

setup_kwargs = {
    'name': 'mlcvzoo-tf-classification',
    'version': '5.0.0',
    'description': 'MLCVZoo TF Classifcation Package',
    'long_description': '# MLCVZoo TF Classification\n\nThe MLCVZoo is an SDK for simplifying the usage of various (machine learning driven)\ncomputer vision algorithms. The package **mlcvzoo_tf_classification** is the wrapper module\nfor classification algorithms that are implemented using tensorflow.\n\nFurther information about the MLCVZoo can be found [here](../README.md).\n\n## Install\n`\npip install mlcvzoo-tf-classification\n`\n\n## Technology stack\n\n- Python\n',
    'author': 'Maximilian Otten',
    'author_email': 'maximilian.otten@iml.fraunhofer.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git.openlogisticsfoundation.org/silicon-economy/base/ml-toolbox/mlcvzoo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
The TrainableNetBasedModel class configures the model for training and inference.
The respective model is compiled from the provided configuration file and net class.
Input parameters are parsed via the AnnotationMapper.
Data flow is regulated via Generators.
"""
import logging
import os
import typing
from typing import Dict, Optional

import numpy as np
import tensorflow as tf
from mlcvzoo_base.configuration.utils import (
    create_configuration as create_basis_configuration,
)
from tensorflow.keras.applications.xception import Xception, preprocess_input

from mlcvzoo_tf_classification.base_model import BaseModel, ImageType
from mlcvzoo_tf_classification.xception.configuration import XceptionConfig

logger = logging.getLogger(__name__)


class XceptionModel(
    BaseModel[XceptionConfig],
):
    """
    The model is compiled from the respective net class and configuration file.
    Input details are parsed using the AnnotationClassMapper.
    """

    def __init__(
        self,
        from_yaml: str,
        configuration: Optional[XceptionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ):
        BaseModel.__init__(
            self,
            from_yaml=from_yaml,
            configuration=configuration,
            string_replacement_map=string_replacement_map,
        )

    def _init_inference_model(self) -> None:
        input_shape_config = tuple(self.configuration.net_config.input_shape)

        weights_config = None
        # check both, as model_path can be either the preconfigured keras weights or
        # pretrained weights at path
        if self.configuration.net_config.model_path == "imagenet" or os.path.isfile(
            self.configuration.net_config.model_path
        ):
            weights_config = self.configuration.net_config.model_path

        self.net = Xception(
            # custom_net implies (not include_top)
            include_top=not self.configuration.net_config.build_custom_net,
            weights=weights_config,
            input_tensor=None,
            input_shape=input_shape_config,
            pooling=None,
            classes=self.mapper.num_classes,
            classifier_activation="softmax",
        )

        if self.configuration.net_config.build_custom_net:
            self.net.trainable = False
            # set input shape
            inputs = tf.keras.Input(shape=self.net.input_shape[1:])

            # set custom top
            previous_net_output = self.net(inputs, training=False)
            in_between_layer = tf.keras.layers.Flatten()(previous_net_output)
            outputs = tf.keras.layers.Dense(self.mapper.num_classes)(in_between_layer)

            self.net = tf.keras.Model(inputs, outputs)

    def store(self, checkpoint_path: str) -> None:
        BaseModel.store(self, checkpoint_path)

    def restore(self, checkpoint_path: str) -> None:
        BaseModel.restore(self, checkpoint_path)

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[XceptionConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> XceptionConfig:
        return typing.cast(
            XceptionConfig,
            create_basis_configuration(
                configuration_class=XceptionConfig,
                from_yaml=from_yaml,
                input_configuration=configuration,
                string_replacement_map=string_replacement_map,
            ),
        )

    @staticmethod
    def preprocess_data(input_data: ImageType) -> ImageType:
        return preprocess_input(input_data)  # type: ignore
